/* MIDI Square
 * original version written in HiSoft BASIC 5/18/93
 * Copyright 1993 David Snow
 *
 * C translation: (c) 4/2001 Edgar Aichinger
 * compiler: 	gcc 2.95
 * libraries:	cflib 0.20, gemlib 0.40, mintlibs 0.57
 *
 * for changes look in new.txt
 */
 
/*******************************************************************/
#include <cflib.h>
#include <mint/osbind.h>
#include <support.h>
#include <time.h>

#include "midisqar.h"

#ifndef FALSE
# define FALSE 0
# define TRUE !FALSE
#endif

#define BUFFER_SIZE 			0x20000
#define STEPS_PER_QUARTERNOTE	96			/* resolution */

#define RSC_NAME 				"midisqar.rsc"
#define PRG_NAME				" MIDISquare "

#define beep() Bconout (2, 7)

#define NOTE_OFF	0x80
#define NOTE_ON		0x90
#define PRG_CHG		0xC0

#define SMF_META_EVENT			0xFF
#define SMF_META_SET_TEMPO		0x51
#define SMF_META_END_OF_TRACK	0x2F

#define midi_is()		Bconstat(_MIDI)
#define midi_in()		Bconin(_MIDI)
#define midi_out(a)		Bconout(_MIDI,a)

typedef char			byte;			/*	 8 bits, signed	*/
typedef short			word;			/*  16 bits, signed	*/
typedef unsigned char	ubyte;			/*   8 bits, unsigned */
typedef unsigned short	uword; 			/*  16 bits, unsigned */
typedef unsigned long	ulong;			/*  32 bits, unsigned */

#define lowerbyte(a)	((ubyte)(a & 0xff))
#define upperbyte(a)	((ubyte)((a >> 8) & 0xff))
#define upmostbyte(a)	((ubyte)((a >> 16) & 0xff))

#define MThd 	0x4D546864
#define MTrk 	0x4D54726B

typedef struct {
	long 	thd_magic;			/*	"MThd" */
	long 	hdr_len;			/*	length of header */
	short 	format;				/* 	SMF format (0 - 2) */
	short 	tracks;				/*	# of tracks */
	short 	division;			/* 	steps per beat (quarter note) */
} HDRCHUNK;

typedef struct {
	long trk_magic;				/*	"MTrk" */
	long trk_len;				/*	sequence length */
} TRKCHUNK;

typedef struct {
	ubyte		a;
	ubyte		b;
	ubyte		c;
} SMFLong;

/*******************************************************************/
/*
 * global variables
 */

extern short _app;

short 	msg[8], menu_id;
//short 	vdi_handle;

short	event, msx, msy, mbutton, kstate, kreturn, mclick;
short 	quit = FALSE;

OBJECT 	*midisqar_pointer;		/* main tree */
OBJECT 	*msq_icon;				/* window icon */

char	*alert_mem,				/* alert strings */
		*alert_exit,
		*alert_openfile,
		*alert_overwrite,
		*alert_discard;

WDIALOG	*wd_main;

short 	square_top, square_bottom,
		square_left, square_right;
short 	x, y, w, h;

short 	x_offset = 4, y_offset = -4,
		x_velocity = 96, y_velocity = 96;

short 	m_channel[4] = {0, 0, 0, 0};
short 	global_channel, global_transposition;
short 	inverse_velocity;
short 	tempo = 120;

short 	idx = 0;

short 	scale_gamut[48];
short 	source_scale[7] = {1,3,5,6,8,10,12};

short 	units_index = 3;
short 	units_per_beat[6] = {1,2,3,4,6,8};
float 	unit_duration;
short 	steps_per_unit;
long 	delta_time = 0;

ubyte 	smf_buffer[BUFFER_SIZE];
short 	smf_index;
short 	midifile_flag = 0;

char 	path[256];
char 	midifile_path[256];
char 	file_name[256] = "new.mid";

HDRCHUNK smf_header = {
	 MThd,						/* "MThd" magic (defined in midilib.h) */
	 6,							/* len */
	 0,							/* format 0 */
	 1,							/* 1 track */
	 STEPS_PER_QUARTERNOTE		/* division */
};

TRKCHUNK track = {
	 MTrk,						/* "MTrk" magic (defined in midilib.h) */
	 0,							/* len */
};

/*******************************************************************/
/*
 * local functions
 */
/*******************************************************************/
/* long -> SMFLong (3 bytes, MSB first) */
static SMFLong 
long_to_i3(long l)
{
	SMFLong data;
	
	data.a = upmostbyte(l);
	data.b = upperbyte(l);
	data.c = lowerbyte(l);
	
	return (data);
}
/*******************************************************************/
/* show a value (always positive) in a window object */
static void 
show_unsigned (short value, short obj)
{
	char text[6];

	_ltoa (value, text, 10);
	set_string(wd_main->tree, obj, text);
	redraw_wdobj(wd_main, obj);
}
/* show a value (positive or negative) in a window object */
static void 
show_signed (short value, short obj)
{
	char text[6];

	sprintf (text, "%+d", value);
	set_string(wd_main->tree, obj, text);
	redraw_wdobj(wd_main, obj);
}

/*******************************************************************/
static void 
select_mode (void)
{
	if (get_state(wd_main->tree, POLY_MODE, OS_SELECTED))
		for (idx = 1; idx <= 3; idx++)
			m_channel[idx] = m_channel[0];
	else 
		for (idx = 1; idx <= 3; idx++)
			m_channel[idx] = (m_channel[idx-1]+1) % 16;
}
static void 
select_velocity (void)
{
	inverse_velocity = !(get_state (wd_main->tree, VELOCITY_A, OS_SELECTED));
}
/*******************************************************************/
/*
 * Write variable-length bytes to MIDI format files
 * taken from Tim Thompson's midifile(3) library
 * (modified for use in this code)
 */
static void 
delta_conversion (void)
{
	unsigned long vlq_buffer;

	vlq_buffer = delta_time & 0x7f;
	while((delta_time >>= 7) > 0) {
		vlq_buffer <<= 8;
		vlq_buffer |= 0x80;
		vlq_buffer += (delta_time & 0x7f);
	}
	while(1) {
		smf_buffer[smf_index] = vlq_buffer & 0xff;
		smf_index++;
		
		if(vlq_buffer & 0x80)
			vlq_buffer >>= 8;
		else
			return;
	}
}
/*******************************************************************/
/*
 * actual key handler
 *	called directly from main() with evnt_multi() return values
 *	and via key_check() from play_square() with Bconin() return values
 *  handles both numpad and keyboard numbers
 */
static void 
handle_keys (char shift, char scancode)
{
	if ((shift & K_LSHIFT) || (shift & K_RSHIFT))
		switch (scancode) {
			case 0x02:
			case 0x6d:	global_transposition = -1;	break;
			case 0x03:
			case 0x6e:	global_transposition = -2;	break;
			case 0x04:
			case 0x6f:	global_transposition = -3;	break;
			case 0x05:
			case 0x6a:	global_transposition = -4;	break;
			case 0x06:
			case 0x6b:	global_transposition = -5;	break;
			case 0x07:
			case 0x6c:	global_transposition = -6;	break;
			case 0x08:
			case 0x67:	global_transposition = -7;	break;
			case 0x09:
			case 0x68:	global_transposition = -8;	break;
			case 0x0a:
			case 0x79:	global_transposition = -9;	break;
			case 0x0b:
			case 0x70:	global_transposition =  0;	break;
		}
	else
		switch (scancode) {
			case 0x02:
			case 0x6d:	global_transposition =  1;	break;
			case 0x03:
			case 0x6e:	global_transposition =  2;	break;
			case 0x04:
			case 0x6f:	global_transposition =  3;	break;
			case 0x05:
			case 0x6a:	global_transposition =  4;	break;
			case 0x06:
			case 0x6b:	global_transposition =  5;	break;
			case 0x07:
			case 0x6c:	global_transposition =  6;	break;
			case 0x08:
			case 0x67:	global_transposition =  7;	break;
			case 0x09:
			case 0x68:	global_transposition =  8;	break;
			case 0x0a:
			case 0x79:	global_transposition =  9;	break;
			case 0x0b:
			case 0x70:	global_transposition =  0;	break;
		}
	show_signed(global_transposition, TRANSPOSITION);
}

/*******************************************************************/
/*
 * wrapper into handle_keys() for play_square()
 */
static void 
key_check (void)
{
	long key = 0;
	char scancode, shift;	
	
	if (Bconstat(_CON)) {
		shift = Kbshift(-1);
		key = Bconin(_CON);
		
		scancode = (char)(key >> 16 & 0xff );
		
		handle_keys (shift, scancode);
	}
}
/*******************************************************************/
static void 
sq_note_off(short global_channel, short pitch)
{
	midi_out (NOTE_OFF + global_channel);
	midi_out (pitch);
	midi_out (64);
	
	if (midifile_flag) {
		delta_conversion();
		smf_buffer[smf_index] = NOTE_OFF + global_channel;
		smf_index++;
		smf_buffer[smf_index] = pitch;
		smf_index++;
		smf_buffer[smf_index] = 64;
		smf_index++;
	}
}

static void 
sq_note_on(short global_channel, short pitch, short velocity)
{
	midi_out (NOTE_ON + global_channel);
	midi_out (pitch);
	midi_out (velocity);
	
	if (midifile_flag) {
		delta_conversion();
		smf_buffer[smf_index] = NOTE_ON + global_channel;
		smf_index++;
		smf_buffer[smf_index] = pitch;
		smf_index++;
		smf_buffer[smf_index] = velocity;
		smf_index++;
	}
}

#if 0
static void 
program_change(short global_channel, short number)
{
	midi_out (PRG_CHG + global_channel);
	midi_out (number);
	
	if (midifile_flag) {
		delta_conversion();
		smf_buffer[smf_index] = PRG_CHG + global_channel;
		smf_index++;
		smf_buffer[smf_index] = number;
		smf_index++;
	}
}
#endif

static void 
fill_gamut (void)
{
	for (idx=0; idx<6; idx++) {
		scale_gamut[idx] = source_scale[idx] + 23;
		scale_gamut[idx+7] = source_scale[idx] + 12 + 23;
		scale_gamut[idx+14] = source_scale[idx] + 24 + 23;
		scale_gamut[idx+21] = source_scale[idx] + 36 + 23;
		scale_gamut[idx+28] = source_scale[idx] + 48 + 23;
		scale_gamut[idx+35] = source_scale[idx] + 60 + 23;
		scale_gamut[idx+42] = source_scale[idx] + 72 + 23;
	}
}
static void 
set_unit_duration (void)
{
	float step_duration = 60 / (float)tempo / STEPS_PER_QUARTERNOTE;
	
	steps_per_unit = STEPS_PER_QUARTERNOTE / units_per_beat[units_index];
	unit_duration = steps_per_unit * step_duration; /* in seconds */
}
static void 
set_tempo (void)
{
	long smf_tempo =60 * 1000000 / tempo;
	SMFLong smft_i3 = long_to_i3(smf_tempo);
	
	smf_buffer[smf_index] = 0x00;	//delta time
	smf_index++;
	smf_buffer[smf_index] = SMF_META_EVENT;	//meta event
	smf_index++;
	smf_buffer[smf_index] = SMF_META_SET_TEMPO;
	smf_index++;
	smf_buffer[smf_index] = 0x03;
	smf_index++;
	smf_buffer[smf_index] = smft_i3.a;
	smf_index++;
	smf_buffer[smf_index] = smft_i3.b;
	smf_index++;
	smf_buffer[smf_index] = smft_i3.c;
	smf_index++;
}
/*******************************************************************/
static void 
play_square (void)
{
	float now, next_unit;					/* in seconds */
	short x_index = 0, y_index = 0;
	short old_x_index = -1, old_y_index = -1;
	short mouse_x = 0, mouse_y = 0, mouse_button = 0, dummy = 0;
	short note_on_flag = 0;
	
	/* update grid coords (in case window has been moved) */
	objc_offset(wd_main->tree, MIDI_GRID, &square_left, &square_top);

	square_bottom = square_top + 256;
	square_right = square_left + 256;

	graf_mouse (6,0);

//	wind_update(BEG_MCTRL);
	next_unit = (float)clock()/CLOCKS_PER_SEC;
	delta_time = 0 - steps_per_unit;		/* initialize */
	do {
		now = (float)clock()/CLOCKS_PER_SEC;
		if (now >= next_unit) {
			next_unit = now + unit_duration;
			delta_time += steps_per_unit;
			
			graf_mkstate (&mouse_x, &mouse_y, &mouse_button, &dummy);
			if (mouse_button == 0) {
				if (note_on_flag) {
					note_on_flag = 0;
					sq_note_off (m_channel[0], scale_gamut[old_x_index] + global_transposition);
					sq_note_off (m_channel[1], scale_gamut[old_x_index + x_offset]+global_transposition);
					sq_note_off (m_channel[2], scale_gamut[old_y_index] + global_transposition);
					sq_note_off (m_channel[3], scale_gamut[old_y_index + y_offset]+global_transposition);
					old_x_index=-1;
					old_y_index=-1;
					key_check();
				}
				else
					key_check();
				if ((mouse_x < square_left) | (mouse_x >= square_right) | 
						(mouse_y < square_top) | (mouse_y >= square_bottom)) {
//					wind_update(END_MCTRL);
					graf_mouse (0,0);
					break;	/* exits do{} while(1); loop */
				}
			}
			else {	/* mouse_button == 1 */
				if ((mouse_x >= square_left) & (mouse_x < square_right) & 
						(mouse_y >= square_top) & (mouse_y < square_bottom)) {
					x_index = (mouse_x - square_left)/8 + 7;
					y_index = (mouse_y - square_top)/8 + 7;
	
					if ((x_index != old_x_index) | (y_index != old_y_index)) {
						if (mouse_button & 2) {
							x_velocity = (x_index-7)*3;
							y_velocity = (y_index-7)*3;
							if (inverse_velocity) {
								x_velocity = 93 - x_velocity;
								y_velocity = 93 - y_velocity;
							}
							x_velocity = x_velocity+32;
							y_velocity = y_velocity+32;
						}
						if (mouse_button & 1) {
							if (note_on_flag) {
								sq_note_off (m_channel[0], scale_gamut[old_x_index] + global_transposition);
								sq_note_off (m_channel[1], scale_gamut[old_x_index + x_offset] + global_transposition);
								sq_note_off (m_channel[2], scale_gamut[old_y_index] + global_transposition);
								sq_note_off (m_channel[3], scale_gamut[old_y_index + y_offset] + global_transposition);
								key_check();
							}
							sq_note_on (m_channel[0], scale_gamut[x_index] + global_transposition, x_velocity);
							sq_note_on (m_channel[1], scale_gamut[x_index + x_offset] + global_transposition, x_velocity);
							sq_note_on (m_channel[2], scale_gamut[y_index] + global_transposition, y_velocity);
							sq_note_on (m_channel[3], scale_gamut[y_index + y_offset] + global_transposition, y_velocity);
							note_on_flag = 1;
							old_x_index = x_index;
							old_y_index = y_index;
						}
					}
				}
			}	/* mouse_button == 1 */
		}
	} while (1);	/* endless loop */
}
/*******************************************************************/
static void 
save_midifile (void)
{
	short ok, smf_done = FALSE;
	int fp;
	
//	beep();

	//APPEND END-OF-TRACK META-EVENT
	delta_time=0;
	delta_conversion();

	smf_buffer[smf_index] = SMF_META_EVENT;	/* meta-event */
	smf_index++;
	smf_buffer[smf_index] = SMF_META_END_OF_TRACK;
	smf_index++;
	smf_buffer[smf_index] = 0x00;
	smf_index++;

	track.trk_len = smf_index;				/* update sequence length */
	
	do {
		ok = select_file (path, file_name, "*.mid", " Save MIDIfile as... ", FSCB_NULL);
		if (ok) {
			if (!file_name) {
				if (do_walert( 1, 1, alert_discard, PRG_NAME) != 2)
					smf_done = TRUE;
			}
			else {
				graf_mouse (2, 0);
				strcpy (midifile_path, path);
				strcat (midifile_path, file_name);
				if (file_exists (midifile_path)) {
					if (do_walert( 1, 1, alert_overwrite, PRG_NAME) != 1)
						Fdelete (midifile_path);
					else
						break;
				}
				
				fp = Fopen (midifile_path,1);
				if (fp < 0) {
					fp = Fcreate (midifile_path, 0);
					if (fp < 0) {
						do_walert (1, 1, alert_openfile, PRG_NAME );
					}
				}
				Fwrite (fp, sizeof(HDRCHUNK), &smf_header);
				Fwrite (fp, sizeof(TRKCHUNK), &track);
				Fwrite (fp, track.trk_len, smf_buffer);
				Fclose (fp);
				graf_mouse (0, 0);
				smf_done = TRUE;
			}
		}
		else
			if (do_walert( 1, 1, alert_discard, PRG_NAME) != 2)
				smf_done = TRUE;
	} while (!smf_done);
}

/*******************************************************************/
/*
 * callbacks for main dialog
 */
static void
main_open_cb (WDIALOG *wd)
{
	show_unsigned(tempo, BEATS_PER_MINUTE);
	show_unsigned(m_channel[0] + 1, MIDI_CHANNEL);
	show_signed(global_transposition, TRANSPOSITION);
	show_signed(x_offset, HORIZ_INTERVAL);
	show_signed(y_offset, VERT_INTERVAL);
	show_unsigned(units_per_beat[units_index], BEAT_DIVISION);
	
//	draw_grid ();

}
/*******************************************************************/
static int
main_exit_cb (WDIALOG *wd, short obj)
{
	short new_tempo_flag = FALSE;
	short close = FALSE;
	
	switch (obj) {
		case BPM_UP:
			if (tempo<240) {
				new_tempo_flag = TRUE;
				tempo++;
				show_unsigned(tempo, BEATS_PER_MINUTE);
			}
			break;
		case BPM_DOWN:
			if (tempo>20) {
				new_tempo_flag = TRUE;
				tempo--;
				show_unsigned(tempo, BEATS_PER_MINUTE);
			}
			break;
		case CHANNEL_UP:
			for (idx = 0; idx <= 3; idx++) {
				m_channel[idx]++;
				if (m_channel[idx] > 15)
					m_channel[idx] = 0;
			}
			show_unsigned(m_channel[0] + 1, MIDI_CHANNEL);
			break;
		case CHANNEL_DOWN:
			for (idx = 0; idx <= 3; idx++) {
				m_channel[idx]--;
				if (m_channel[idx] < 0)
					m_channel[idx] = 15;
			}
			show_unsigned(m_channel[0] + 1, MIDI_CHANNEL);
			break;
		case TRANSPOSE_UP:
			if (global_transposition < 9)
				global_transposition++;
			show_signed(global_transposition, TRANSPOSITION);
			break;
		case TRANSPOSE_DOWN:
			if (global_transposition > -9)
				global_transposition--;
			show_signed(global_transposition, TRANSPOSITION);
			break;
		case HORIZONTAL_UP:
			if (x_offset<7)
				x_offset++;
			show_signed(x_offset, HORIZ_INTERVAL);
			break;
		case HORIZONTAL_DOWN:
			if (x_offset>-7)
				x_offset--;
			show_signed(x_offset, HORIZ_INTERVAL);
			break;
		case VERTICAL_UP:
			if (y_offset<7)
				y_offset++;
			show_signed(y_offset, VERT_INTERVAL);
			break;
		case VERTICAL_DOWN:
			if (y_offset>-7)
				y_offset--;
			show_signed(y_offset, VERT_INTERVAL);
			break;
		case UNIT_UP:
			if (units_index<5)
				units_index++;
			show_unsigned(units_per_beat[units_index], BEAT_DIVISION);
			break;
		case UNIT_DOWN:
			if (units_index > 0)
				units_index--;
			show_unsigned(units_per_beat[units_index], BEAT_DIVISION);
			break;
		case START_BUTTON:
			if (midifile_flag == FALSE) {	
//				beep();
				midifile_flag = TRUE;
				track.trk_len = 0;			/* dummy track length */
				smf_index = 0;				/* was 22 in basic version (header length)*/
				set_tempo();
			}
			break;
		case STOP_BUTTON:
			if (midifile_flag) {
//				beep();
				midifile_flag=0;
				save_midifile();
			}
			break;
		case MIDI_GRID:
			select_mode();
			select_velocity();
			set_unit_duration();
			if (midifile_flag & new_tempo_flag) {
				new_tempo_flag = FALSE;
				set_tempo();
			}
			play_square();
			break;
		case C_NATURAL:	source_scale[0]=1;	fill_gamut();	break;
		case C_FLAT:	source_scale[0]=0;	fill_gamut();	break;
		case C_SHARP:	source_scale[0]=2;	fill_gamut();	break;
		case D_NATURAL:	source_scale[1]=3;	fill_gamut();	break;
		case D_FLAT:	source_scale[1]=2;	fill_gamut();	break;
		case D_SHARP:	source_scale[1]=4;	fill_gamut();	break;
		case E_NATURAL:	source_scale[2]=5;	fill_gamut();	break;
		case E_FLAT:	source_scale[2]=4;	fill_gamut();	break;
		case E_SHARP:	source_scale[2]=6;	fill_gamut();	break;
		case F_NATURAL:	source_scale[3]=6;	fill_gamut();	break;
		case F_FLAT:	source_scale[3]=5;	fill_gamut();	break;
		case F_SHARP:	source_scale[3]=7;	fill_gamut();	break;
		case G_NATURAL:	source_scale[4]=8;	fill_gamut();	break;
		case G_FLAT:	source_scale[4]=7;	fill_gamut();	break;
		case G_SHARP:	source_scale[4]=9;	fill_gamut();	break;
		case A_NATURAL:	source_scale[5]=10;	fill_gamut();	break;
		case A_FLAT:	source_scale[5]=9;	fill_gamut();	break;
		case A_SHARP:	source_scale[5]=11;	fill_gamut();	break;
		case B_NATURAL:	source_scale[6]=12;	fill_gamut();	break;
		case B_FLAT:	source_scale[6]=11;	fill_gamut();	break;
		case B_SHARP:	source_scale[6]=13;	fill_gamut();	break;
		case WD_CLOSER:
		case EXIT_BUTTON:
			if (_app) {
				if (do_walert( 1, 1, alert_exit, PRG_NAME) == 1) {
					close = TRUE;
					quit = TRUE;
				}
				else
					set_state(wd_main->tree, EXIT_BUTTON, OS_SELECTED, 0);
			}
			else
				close = TRUE;
			break;
	}
	if ((obj != MIDI_GRID) & (!close))
		redraw_wdobj (wd_main, obj);
	return (close);
}
/*******************************************************************/
static void
handle_msg (short *msg)
{
	if (!message_wdial (msg))
	{
		switch (msg[0])
		{
			case MN_SELECTED:
				break;
			case WM_CLOSED:
			case WM_BOTTOMED:
			case WM_REDRAW:
			case WM_NEWTOP:
			case WM_TOPPED:
			case WM_MOVED:
			case WM_SIZED:
				break;
			case AC_OPEN:
				if (msg[4] == menu_id) {
					wd_main = create_wdial (midisqar_pointer, msq_icon, 0, 
						main_open_cb, main_exit_cb);
					open_wdial (wd_main, -1, -1);
				}
				break;
			case AC_CLOSE:
				if (msg[4] == menu_id) {
					close_wdial (wd_main);
					delete_wdial (wd_main);
				}
				break;
			case AP_TERM:
				quit = TRUE;
				break;
		}
	}
}
/*******************************************************************/
static int 
init_rsc (void)
{
	if ( rsrc_load( RSC_NAME ) == 0 )
	{								/* the only untranslatable alert :( */
		do_walert( 1, 1, "[1][midisqar.rsc missing.][Abort]", PRG_NAME );
		return (FALSE);
	}
	else {
		rsrc_gaddr( R_TREE, MAIN_SCREEN, &midisqar_pointer);
		fix_dial (midisqar_pointer);
		
		rsrc_gaddr (R_TREE, MSQ_ICON, &msq_icon);
		if (_app)
			wd_main = create_wdial (midisqar_pointer, msq_icon, 0, 
					main_open_cb, main_exit_cb);
		
		rsrc_gaddr( R_STRING, ALERT_MEM, &alert_mem );
		rsrc_gaddr( R_STRING, ALERT_EXIT, &alert_exit );
		rsrc_gaddr( R_STRING, ALERT_OPENFILE, &alert_openfile );
		rsrc_gaddr( R_STRING, ALERT_OVERWRITE, &alert_overwrite );
		rsrc_gaddr( R_STRING, ALERT_DISCARD, &alert_discard );
		
		return (TRUE);
	}
}
/*******************************************************************/
/*
 * main()
 */
int 
main (void)
{
	init_app (RSC_NAME);
	if ((!_app) || (gl_ap_version >= 0x410))
		menu_id = menu_register (gl_apid, "  MIDISquare ");

	graf_mouse (256,0);
	
	if (init_rsc()) {	
		
		/* get current drive/path */
		get_path(path, 0);
		
		fill_gamut();
		
		/* callback for window alerts etc. (for correct redraws) */
		set_mdial_wincb (handle_msg);
	
		if (_app) {
			open_wdial (wd_main, -1, -1);
		}
		
//		beep();
		graf_mouse (257,0);
		
		do {
			mbutton = 0;
			event = evnt_multi (MU_MESAG | MU_BUTTON | MU_KEYBD | MU_TIMER,
						 1, 1, 1,
						 0, 0, 0, 0, 0,
						 0, 0, 0, 0, 0,
						 msg,
						 1,
						 &msx, &msy, &mbutton, &kstate,
						 &kreturn, &mclick);
		
			if (event & MU_MESAG)
				handle_msg (msg);
		
			if (event & MU_BUTTON)
			{
				if (!click_wdial (mclick, msx, msy, kstate, mbutton))
					;
			}
			if (event & MU_KEYBD)
			{
				if (!key_wdial (kreturn, kstate))
					printf ("%c, %c\n",(char)kstate, (char)(kreturn >> 8));
					handle_keys ((char)kstate, (char)(kreturn >> 8));
			}
			if (event & MU_TIMER)
				;
		} while (!quit);
		
	}
	if (_app)	
		delete_wdial (wd_main);
	exit_gem();
	return (FALSE);
}

